/* Datei: tss.c */
#include <stdio.h>
#include <stdlib.h>   /* malloc() */
#include <threads.h>
#include <stdbool.h>

/* Terminierungsvariable (vereinfacht, nicht synchronisiert)*/
int terminate = false;

/* globaler Schluessel fuer TSS */
tss_t tsskey;

/* Datenstruktur des TSS */
typedef struct 
{
   unsigned char data;
} threadLocalData_t;

/* Ausgabefunktion fuer TSS-Daten */
void evaluate (void) 
{
   threadLocalData_t * tld = (threadLocalData_t*)tss_get (tsskey);
   printf ("data = %03i, Thread 0x%08x\n", tld->data
           , (int) thrd_current().p);
}

/* Thread Funktion */
int threadfunction (void * arg)
{
   /* Initialwert fuer Threadinstanz auslesen */
   unsigned char inival = * (unsigned char *) arg;
   threadLocalData_t * pLocal = malloc (sizeof (threadLocalData_t));

   /* TSS mit dem globalen Schluessel verknuepfen */
   tss_set (tsskey, pLocal);

   /* Initialwert fuer aktuellen Thread zuweisen */
   pLocal->data = inival;
   while (true)
   {
       /* lokale Daten inkrementieren und Evaluation aufrufen */
       pLocal->data++;
       evaluate();

       /* abfragen, ob Thread beendet werden soll */
       if (terminate == true)
       {
           break;
       }
       thrd_yield();
   }

   /* lokale Daten wieder freigeben */
   free (pLocal);
   return thrd_success;
}

int main (void) 
{
   /* Variable zum Speichern der Thread ID */
   thrd_t thr1, thr2;

   /* Variablen fuer Initialwerte */
   unsigned char iniVal1 = 0;
   unsigned char iniVal2 = 128;

   /* Initialisierung des TSS-Schluessels und */
   /* Erzeugung der Threads */
   tss_create (&tsskey, NULL);
   thrd_create (&thr1, threadfunction, (void *) &iniVal1);
   thrd_create (&thr2, threadfunction, (void *) &iniVal2);

   for (int i = 0; i < 1e8; i++) 
   {
       /* eine Weile warten... */
   }

   /* Beenden der Threads und Freigeben des TSS-Schluessels */
   terminate = true;
   thrd_join (thr1, NULL);
   thrd_join (thr2, NULL);
   tss_delete (tsskey);
}